#!/usr/bin/perl
# $Id: play5.pl 369 2017-02-25 16:01:02Z bd $

#
# This script calls dvbv5-zap, and does some bookkeeping.
#

use Time::HiRes qw(sleep);

my $PATHPREFIX = "/usr/local/glotv";
my $zap = "/usr/bin/dvbv5-zap";
my $conf = "/usr/local/glotv/etc/dvb5-channels.conf";
my $dvr_device = "/dev/dvb/adapter0/dvr0";
my $ffmpeg_fmt = "ffmpeg -i %s -map 0:#%d? -map 0:#%d? -nostats -loglevel fatal -vcodec copy -acodec copy -f mpegts pipe:1";
my $play_pid = $$;
my $zap_pid;
my $ffmpeg_pid;
my $restart;
my $mpegbufsize = 32768;
my $frequency = undef;
my $video_pid = undef;
my $audio_pid = undef;
my $delivery_system = undef;

{
    open CONF, "<$conf" or die "Couldn't open configuration file $conf:$!";
    my %params = ();
    my $current_key = undef;
    my $current_map = {};
    while (defined(my $line=<CONF>)) {
	chomp $line;
	if ($line =~ m/\[(.+)\]/) {
	    $current_key = $1;
	    $current_map = {};
	    $params{$current_key} = $current_map;
	} else {
	    if ($line =~ /([A-Z_]+) = (\S+)/) {
		$$current_map{$1} = $2;
	    }
	}
    }
    close CONF;
    my $m = $params{&get_requested_channel};
    $frequency = $$m{FREQUENCY};
    $video_pid = $$m{VIDEO_PID};
    $audio_pid = (split / /, $$m{AUDIO_PID})[0];
    $delivery_system = $$m{DELIVERY_SYSTEM};
}

sub cleanup_and_exit {
    &cleanup_children();
    unlink "$PATHPREFIX/var/dvb-play.pid";
    exit(0);
}

sub cleanup_children {
    #print STDERR "cleanup_children $ffmpeg_pid $zap_pid\n";
    kill 'SIGKILL', $ffmpeg_pid if $ffmpeg_pid;
    kill 'SIGTERM', $zap_pid if $zap_pid;
    unlink "$PATHPREFIX/var/dvb-playing-channel";
    system "pkill dvbv5-zap";  # Temporary workaround for strange zap issue...
}

sub terminate_children {
    kill 'SIGKILL', $ffmpeg_pid if $ffmpeg_pid;
    kill 'SIGTERM', $zap_pid;
    exit(0);
}

sub get_requested_channel {
    open CHAN, '<', "$PATHPREFIX/var/dvb-requested-channel"
	or die "Couldn't read $PATHPREFIX/var/dvb-requested-channel: $!\n";
    my $chan = <CHAN>;
    close CHAN;
    chomp $chan;
    return $chan;
}

$SIG{PIPE} = 'IGNORE';
$SIG{CHLD} = 'IGNORE';

if (open PID, '<', "$PATHPREFIX/var/dvb-play.pid") {
    my $pid = <PID>;
    close PID;
    chomp $pid;
    if ($pid) {
	kill 'SIGTERM', $pid;
	sleep 0.1;
    }
}
open PID, '>', "$PATHPREFIX/var/dvb-play.pid"
    or die "Couldn't write to $PATHPREFIX/var/dvb-play.pid: $!";
print PID $play_pid,"\n";
close PID;

while (1) {
    my $chan = &get_requested_channel();
    my $giveup_counter = 10;
    while ($giveup_counter > 0) {
	system("dvb-fe-tool -d $delivery_system");
	$zap_pid = open ZAP, '-|', "$zap -P -r -c $conf '$frequency' 2>&1";
	die "Couldn't run $zap" unless defined $zap_pid;
	$SIG{TERM} = \&cleanup_and_exit;
	$SIG{HUP} = sub { $restart=1; };
	$restart = 0;
	my $nosignal_counter = 2;
	while ($nosignal_counter > 0) {
	    my $line = <ZAP>;
	    exit(1) unless defined $line;
	    print STDERR $line;
	    last if $line =~ m/DVR interface '(.+)' can now be opened/;
	    if ($line =~ m/       \(0x00\) Signal=/) {
		$nosignal_counter--;
	    }
	}
	if ($nosignal_counter == 0) {
	    print STDERR "Couldn't get a signal, retrying.\n";
	    &cleanup_children();
	    $giveup_counter--;
	    next;
	}
	if ($giveup_counter == 0) {
	    &cleanup_and_exit();
	    exit(0);	# Safety bumper
	}
	last;
    }
    my $monitor_pid = fork();
    if ($monitor_pid == 0) {
	$SIG{HUP} = \&terminate_children;
	while (1) {
	    my $line = <ZAP>;
	    last unless defined $line;
	    next if $line =~ m/^$/;
	    if ($line !~ m/Lock   \(0x1f\) Signal=/) {
		print STDERR "Houston, we have a problem:\n$line";
		print STDERR "Bad DVB signal status, giving up.\n";
		last;
	    }
	}
	kill 'SIGTERM', $play_pid;
	exit(1);
    } else {
	open SCHAN, '>', "$PATHPREFIX/var/dvb-playing-channel"
	    or die "Couldn't write to $PATHPREFIX/var/dvb-playing-channel: $!";
	print SCHAN $chan,"\n";
	close SCHAN;
	#
	# Copy the video stream
	#
	my $ffmpeg_cmd = sprintf($ffmpeg_fmt, $dvr_device, $video_pid, $audio_pid);
	print STDERR $ffmpeg_cmd, "\n";
	$ffmpeg_pid = open VIDEO, '-|', "exec $ffmpeg_cmd"
	    or die "Couldn't start ffmpeg command $ffmpeg_cmd: $!";
	while (!$restart) {
	    my $iobuf = "";
	    my $n = sysread(VIDEO, $iobuf, $mpegbufsize);
	    if ($n) {
		syswrite(STDOUT, $iobuf, $n);
	    } else {
		&cleanup_children();
		exit(1);
	    }
	}
	close VIDEO;
	print STDERR "Closed the DVB capture...\n";
    }
    kill 'SIGHUP', $monitor_pid;
    unlink "$PATHPREFIX/var/dvb-playing-channel";
    sleep 10.1;
}

